# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Steps for building formula engines with."""

from __future__ import annotations

import logging
import math
from abc import ABC, abstractmethod
from typing import Generic

from frequenz.channels import Receiver, ReceiverError, ReceiverStoppedError

from .._base_types import QuantityT, Sample

_logger = logging.getLogger(__name__)


class FormulaStep(ABC):
    """Represents an individual step/stage in a formula.

    Each step, when applied on to an evaluation stack, would pop its input parameters
    from the stack and push its result back in.
    """

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """

    @abstractmethod
    def apply(self, eval_stack: list[float]) -> None:
        """Apply a formula operation on the eval_stack.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """


class Adder(FormulaStep):
    """A formula step for adding two values."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "+"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, add them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 + val2
        eval_stack.append(res)


class Subtractor(FormulaStep):
    """A formula step for subtracting one value from another."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "-"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, subtract them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 - val2
        eval_stack.append(res)


class Multiplier(FormulaStep):
    """A formula step for multiplying two values."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "*"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, multiply them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 * val2
        eval_stack.append(res)


class Divider(FormulaStep):
    """A formula step for dividing one value by another."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "/"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, divide them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 / val2
        eval_stack.append(res)


class Maximizer(FormulaStep):
    """A formula step that represents the max function."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "max"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack and pushes back the maximum.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = max(val1, val2)
        eval_stack.append(res)


class Minimizer(FormulaStep):
    """A formula step that represents the min function."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "min"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack and pushes back the minimum.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = min(val1, val2)
        eval_stack.append(res)


class Consumption(FormulaStep):
    """A formula step that represents the consumption operator.

    The consumption operator is the maximum of the value on top
    of the evaluation stack and 0.
    """

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "consumption"

    def apply(self, eval_stack: list[float]) -> None:
        """
        Apply the consumption formula.

        Replace the top of the eval eval_stack with the same value if the value
        is positive or 0.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val = eval_stack.pop()
        eval_stack.append(max(val, 0))


class Production(FormulaStep):
    """A formula step that represents the production operator.

    The production operator is the maximum of the value times minus one on top
    of the evaluation stack and 0.
    """

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "production"

    def apply(self, eval_stack: list[float]) -> None:
        """
        Apply the production formula.

        Replace the top of the eval eval_stack with its absolute value if the
        value is negative or 0.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val = eval_stack.pop()
        eval_stack.append(max(-val, 0))


class OpenParen(FormulaStep):
    """A no-op formula step used while building a prefix formula engine.

    Any OpenParen steps would get removed once a formula is built.
    """

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "("

    def apply(self, _: list[float]) -> None:
        """No-op."""


class ConstantValue(FormulaStep):
    """A formula step for inserting a constant value."""

    def __init__(self, value: float) -> None:
        """Create a `ConstantValue` instance.

        Args:
            value: The constant value.
        """
        self._value = value

    @property
    def value(self) -> float:
        """Return the constant value.

        Returns:
            The constant value.
        """
        return self._value

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return str(self._value)

    def apply(self, eval_stack: list[float]) -> None:
        """Push the constant value to the eval_stack.

        Args:
            eval_stack: An evaluation stack, to append the constant value to.
        """
        eval_stack.append(self._value)


class Clipper(FormulaStep):
    """A formula step for clipping a value between a minimum and maximum."""

    def __init__(self, min_val: float | None, max_val: float | None) -> None:
        """Create a `Clipper` instance.

        Args:
            min_val: The minimum value.
            max_val: The maximum value.
        """
        self._min_val = min_val
        self._max_val = max_val

    @property
    def min_value(self) -> float | None:
        """Return the minimum value.

        Returns:
            The minimum value.
        """
        return self._min_val

    @property
    def max_value(self) -> float | None:
        """Return the maximum value.

        Returns:
            The maximum value.
        """
        return self._max_val

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return f"clip({self._min_val}, {self._max_val})"

    def apply(self, eval_stack: list[float]) -> None:
        """Clip the value at the top of the eval_stack.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val = eval_stack.pop()
        if self._min_val is not None:
            val = max(val, self._min_val)
        if self._max_val is not None:
            val = min(val, self._max_val)
        eval_stack.append(val)


class FallbackMetricFetcher(Receiver[Sample[QuantityT]], Generic[QuantityT]):
    """A fallback metric fetcher for formula engines.

    Generates a metric value from the fallback components if the primary metric
    is invalid.

    This class starts running when the primary MetricFetcher starts receiving invalid data.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Get the name of the fetcher."""

    @property
    @abstractmethod
    def is_running(self) -> bool:
        """Check whether the metric fetcher is running."""

    @abstractmethod
    def start(self) -> None:
        """Initialize the metric fetcher and start fetching samples."""

    @abstractmethod
    async def stop(self) -> None:
        """Stope the fetcher if is running."""


class MetricFetcher(Generic[QuantityT], FormulaStep):
    """A formula step for fetching a value from a metric Receiver."""

    def __init__(
        self,
        name: str,
        stream: Receiver[Sample[QuantityT]],
        *,
        nones_are_zeros: bool,
        fallback: FallbackMetricFetcher[QuantityT] | None = None,
    ) -> None:
        """Create a `MetricFetcher` instance.

        Args:
            name: The name of the metric.
            stream: A channel receiver from which to fetch samples.
            nones_are_zeros: Whether to treat None values from the stream as 0s.
            fallback: Metric fetcher to use if primary one start sending
                invalid data (e.g. due to a component stop). If None, the data from
                primary metric fetcher will be used.
        """
        self._name = name
        self._stream: Receiver[Sample[QuantityT]] = stream
        self._next_value: Sample[QuantityT] | None = None
        self._nones_are_zeros = nones_are_zeros
        self._fallback: FallbackMetricFetcher[QuantityT] | None = fallback
        self._latest_fallback_sample: Sample[QuantityT] | None = None
        self._is_stopped = False

    @property
    def stream(self) -> Receiver[Sample[QuantityT]]:
        """Return the stream from which to fetch values.

        Returns:
            The stream from which to fetch values.
        """
        return self._stream

    async def stop(self) -> None:
        """Stop metric fetcher.

        If metric fetcher is stopped, it can't be started again.
        There is no use-case now to start it again.
        """
        self._is_stopped = True
        self.stream.close()
        if self._fallback:
            await self._fallback.stop()

    def stream_name(self) -> str:
        """Return the name of the stream.

        Returns:
            The name of the stream.
        """
        return str(self._stream.__doc__)

    def _is_value_valid(self, value: QuantityT | None) -> bool:
        return not (value is None or value.isnan() or value.isinf())

    async def _fetch_from_fallback(
        self, fallback_fetcher: FallbackMetricFetcher[QuantityT]
    ) -> Sample[QuantityT] | None:
        try:
            return await fallback_fetcher.receive()
        except ReceiverStoppedError:
            if self._is_stopped:
                _logger.debug(
                    "Stream for fallback metric fetcher %s closed.",
                    fallback_fetcher.name,
                )
            else:
                _logger.error(
                    "Failed to fetch next value from %s. Fallback stream closed.",
                    self._name,
                )
            return None
        except ReceiverError as err:
            _logger.error(
                "Failed to fetch next value from fallback stream %s: %s",
                self._name,
                err,
            )
            return None

    async def _synchronize_and_fetch_fallback(
        self,
        primary_fetcher_value: Sample[QuantityT] | None,
        fallback_fetcher: FallbackMetricFetcher[QuantityT],
    ) -> Sample[QuantityT] | None:
        """Synchronize the fallback fetcher and return the fallback value.

        Args:
            primary_fetcher_value: The sample fetched from the primary fetcher.
            fallback_fetcher: The fallback metric fetcher.

        Returns:
            The value from the synchronized stream. Returns None if the primary
            fetcher sample is older than the latest sample from the fallback
            fetcher or if the fallback fetcher fails to fetch the next value.
        """
        # We need to save value, because
        # primary_fetcher_value.timestamp < self._latest_fallback_sample.timestamp
        # In that case we should wait for our time window.
        if self._latest_fallback_sample is None:
            self._latest_fallback_sample = await self._fetch_from_fallback(
                fallback_fetcher
            )

        if primary_fetcher_value is None or self._latest_fallback_sample is None:
            return self._latest_fallback_sample

        if primary_fetcher_value.timestamp < self._latest_fallback_sample.timestamp:
            return None

        # Synchronize the fallback fetcher with primary one
        while primary_fetcher_value.timestamp > self._latest_fallback_sample.timestamp:
            self._latest_fallback_sample = await self._fetch_from_fallback(
                fallback_fetcher
            )
            if self._latest_fallback_sample is None:
                break
        return self._latest_fallback_sample

    async def fetch_next(self) -> Sample[QuantityT] | None:
        """Fetch the next value from the stream.

        To be called before each call to `apply`.

        Returns:
            The fetched Sample.
        """
        if self._is_stopped:
            _logger.error(
                "Metric fetcher %s stopped. Can't fetch new value.", self._name
            )
            return None

        self._next_value = await self._fetch_next()
        return self._next_value

    async def _fetch_next(self) -> Sample[QuantityT] | None:
        # First fetch from primary stream
        primary_value: Sample[QuantityT] | None = None
        try:
            primary_value = await self._stream.receive()
        except ReceiverStoppedError:
            if self._is_stopped:
                _logger.debug("Stream for metric fetcher %s closed.", self._name)
                return None
            _logger.error(
                "Failed to fetch next value from %s. Primary stream closed.",
                self._name,
            )
        except ReceiverError as err:
            _logger.error("Failed to fetch next value from %s: %s", self._name, err)

        # We have no fallback, so we just return primary value even if it is not correct.
        if self._fallback is None:
            return primary_value

        is_primary_value_valid = primary_value is not None and self._is_value_valid(
            primary_value.value
        )

        if is_primary_value_valid:
            # Primary stream is good again, so we can stop fallback and return primary_value.
            if self._fallback.is_running:
                _logger.info(
                    "Primary metric %s is good again, stopping fallback metric fetcher %s",
                    self._name,
                    self._fallback.name,
                )
                await self._fallback.stop()
            return primary_value

        if not self._fallback.is_running:
            _logger.warning(
                "Primary metric %s is invalid. Running fallback metric fetcher: %s",
                self._name,
                self._fallback.name,
            )
            # We started fallback, but it has to subscribe.
            # We will receive fallback values since the next time window.
            self._fallback.start()
            return primary_value

        return await self._synchronize_and_fetch_fallback(primary_value, self._fallback)

    @property
    def value(self) -> Sample[QuantityT] | None:
        """Get the next value in the stream.

        Returns:
            Next value in the stream.
        """
        return self._next_value

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return self._name

    def apply(self, eval_stack: list[float]) -> None:
        """Push the latest value from the stream into the evaluation stack.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.

        Raises:
            RuntimeError: No next value available to append.
        """
        if self._next_value is None:
            raise RuntimeError("No next value available to append.")

        next_value = self._next_value.value
        if next_value is None or next_value.isnan() or next_value.isinf():
            if self._nones_are_zeros:
                eval_stack.append(0.0)
            else:
                eval_stack.append(math.nan)
        else:
            eval_stack.append(next_value.base_value)



# License: MIT
# Copyright © 2025 Frequenz Energy-as-a-Service GmbH

"""A power manager implementation that uses the shifting matryoshka algorithm."""

from __future__ import annotations

import logging
import typing
from datetime import timedelta

from frequenz.quantities import Power
from typing_extensions import override

from frequenz.sdk.timeseries._base_types import Bounds

from ... import timeseries
from . import _bounds
from ._base_classes import BaseAlgorithm, Proposal, _Report

if typing.TYPE_CHECKING:
    from ...timeseries._base_types import SystemBounds

_logger = logging.getLogger(__name__)


def _get_nearest_possible_power(
    power: Power,
    lower_bound: Power,
    upper_bound: Power,
    exclusion_bounds: Bounds[Power] | None,
) -> Power:
    match _bounds.clamp_to_bounds(
        power,
        lower_bound,
        upper_bound,
        exclusion_bounds,
    ):
        case (None, p) | (p, None) if p:
            return p
        case (low, high) if low and high:
            if high - power < power - low:
                return high
            return low
        case _:
            return Power.zero()


class ShiftingMatryoshka(BaseAlgorithm):
    """The ShiftingMatryoshka algorithm.

    When there are multiple actors trying to control the same set of components, this
    algorithm will reconcile the different proposals and calculate the target power.

    Details about the algorithm can be found in the [microgrid module documentation](https://frequenz-floss.github.io/frequenz-sdk-python/v1.0-dev/user-guide/microgrid-concepts/#frequenz.sdk.microgrid--setting-power).
    """  # noqa: E501 (line too long)

    def __init__(self, max_proposal_age: timedelta) -> None:
        """Create a new instance of the matryoshka algorithm."""
        self._max_proposal_age_sec = max_proposal_age.total_seconds()
        self._component_buckets: dict[frozenset[int], set[Proposal]] = {}
        self._target_power: dict[frozenset[int], Power] = {}

    def _calc_targets(
        self,
        component_ids: frozenset[int],
        system_bounds: SystemBounds,
        priority: int | None = None,
    ) -> tuple[Power | None, Bounds[Power]]:
        """Calculate the target power and bounds for the given components.

        Args:
            component_ids: The component IDs to calculate the target power for.
            system_bounds: The system bounds for the components in the proposal.
            priority: The priority of the actor for which the target power is calculated.

        Returns:
            The new target power and bounds for the components.
        """
        proposals = self._component_buckets.get(component_ids, set())
        lower_bound = (
            system_bounds.inclusion_bounds.lower
            if system_bounds.inclusion_bounds
            # if a target power exists from a previous proposal, and the system bounds
            # have become unavailable, force the target power to be zero, by narrowing
            # the bounds to zero.
            else Power.zero()
        )
        upper_bound = (
            system_bounds.inclusion_bounds.upper
            if system_bounds.inclusion_bounds
            else Power.zero()
        )

        if not proposals:
            return None, Bounds[Power](lower=lower_bound, upper=upper_bound)

        available_bounds = Bounds[Power](lower=lower_bound, upper=upper_bound)
        top_pri_bounds: Bounds[Power] | None = None

        target_power = Power.zero()

        allocations: dict[str, str] = {}

        for next_proposal in sorted(proposals, reverse=True):
            # if a priority is given, the bounds calculated until that priority is
            # reached will be the bounds available to an actor with the given priority.
            #
            # This could mean that the calculated target power is incorrect and should
            # not be used.
            if priority is not None and next_proposal.priority <= priority:
                break

            # When the upper bound is less than the lower bound, if means that there's
            # no more room to process further proposals, so we break out of the loop.
            if upper_bound < lower_bound:
                break

            proposal_lower = next_proposal.bounds.lower or lower_bound
            proposal_upper = next_proposal.bounds.upper or upper_bound
            proposal_power = next_proposal.preferred_power

            # Make sure that if the proposal specified bounds, they make sense.
            if proposal_upper < proposal_lower:
                continue

            # If the proposal bounds are outside the available bounds, we need to
            # adjust the proposal bounds to fit within the available bounds.
            if proposal_lower >= upper_bound:
                proposal_lower = upper_bound
                proposal_upper = upper_bound
            elif proposal_upper <= lower_bound:
                proposal_lower = lower_bound
                proposal_upper = lower_bound

            # Clamp the available bounds by the proposal bounds.
            lower_bound = max(lower_bound, proposal_lower)
            upper_bound = min(upper_bound, proposal_upper)

            if proposal_power is not None:
                # If this is the first power setting proposal, then hold on to the
                # bounds that were available at that time, for use when applying the
                # exclusion bounds to the target power at the end.
                if top_pri_bounds is None and proposal_power != Power.zero():
                    top_pri_bounds = Bounds[Power](lower=lower_bound, upper=upper_bound)
                # Clamp the proposal power to its available bounds.
                proposal_power = _get_nearest_possible_power(
                    proposal_power,
                    lower_bound,
                    upper_bound,
                    None,
                )
                # Shift the available bounds by the proposal power.
                lower_bound = lower_bound - proposal_power
                upper_bound = upper_bound - proposal_power
                # Add the proposal power to the target power (aka shift in the opposite direction).
                target_power += proposal_power

                allocations[next_proposal.source_id] = str(proposal_power)

        # The `top_pri_bounds` is to ensure that when applying the exclusion bounds to
        # the target power at the end, we respect the bounds that were set by the first
        # power-proposing actor.
        if top_pri_bounds is not None:
            available_bounds = top_pri_bounds

        # Apply the exclusion bounds to the target power.
        target_power = _get_nearest_possible_power(
            target_power,
            available_bounds.lower,
            available_bounds.upper,
            system_bounds.exclusion_bounds,
        )

        if allocations:
            _logger.info(
                "PowerManager allocations for component IDs: %s: %s",
                sorted(component_ids),
                allocations,
            )

        return target_power, Bounds[Power](lower=lower_bound, upper=upper_bound)

    def _validate_component_ids(
        self,
        component_ids: frozenset[int],
        proposal: Proposal | None,
        system_bounds: SystemBounds,
    ) -> bool:
        if component_ids not in self._component_buckets:
            # if there are no previous proposals and there are no system bounds, then
            # don't calculate a target power and fail the validation.
            if (
                system_bounds.inclusion_bounds is None
                and system_bounds.exclusion_bounds is None
            ):
                if proposal is not None:
                    _logger.warning(
                        "PowerManagingActor: No system bounds available for component "
                        + "IDs %s, but a proposal was given.  The proposal will be "
                        + "ignored.",
                        component_ids,
                    )
                return False

            for bucket in self._component_buckets:
                if any(component_id in bucket for component_id in component_ids):
                    raise NotImplementedError(
                        f"PowerManagingActor: component IDs {component_ids} are already"
                        + " part of another bucket.  Overlapping buckets are not"
                        + " yet supported."
                    )
        return True

    @override
    def calculate_target_power(
        self,
        component_ids: frozenset[int],
        proposal: Proposal | None,
        system_bounds: SystemBounds,
        must_return_power: bool = False,
    ) -> Power | None:
        """Calculate and return the target power for the given components.

        Args:
            component_ids: The component IDs to calculate the target power for.
            proposal: If given, the proposal to added to the bucket, before the target
                power is calculated.
            system_bounds: The system bounds for the components in the proposal.
            must_return_power: If `True`, the algorithm must return a target power,
                even if it hasn't changed since the last call.

        Returns:
            The new target power for the components, or `None` if the target power
                didn't change.

        Raises:  # noqa: DOC502
            NotImplementedError: When the proposal contains component IDs that are
                already part of another bucket.
        """
        if not self._validate_component_ids(component_ids, proposal, system_bounds):
            return None

        if proposal is not None:
            bucket = self._component_buckets.setdefault(component_ids, set())
            if proposal in bucket:
                bucket.remove(proposal)
            if (
                proposal.preferred_power is not None
                or proposal.bounds.lower is not None
                or proposal.bounds.upper is not None
            ):
                bucket.add(proposal)
            elif not bucket:
                del self._component_buckets[component_ids]
                _ = self._target_power.pop(component_ids, None)

        target_power, _ = self._calc_targets(component_ids, system_bounds)

        if target_power is not None and (
            must_return_power
            or component_ids not in self._target_power
            or self._target_power[component_ids] != target_power
        ):
            self._target_power[component_ids] = target_power
            return target_power
        return None

    @override
    def get_status(  # pylint: disable=too-many-locals
        self,
        component_ids: frozenset[int],
        priority: int,
        system_bounds: SystemBounds,
    ) -> _Report:
        """Get the bounds for the algorithm.

        Args:
            component_ids: The IDs of the components to get the bounds for.
            priority: The priority of the actor for which the bounds are requested.
            system_bounds: The system bounds for the components.

        Returns:
            The target power and the available bounds for the given components, for
                the given priority.
        """
        target_power = self._target_power.get(component_ids)
        _, bounds = self._calc_targets(component_ids, system_bounds, priority)
        return _Report(
            target_power=target_power,
            _inclusion_bounds=timeseries.Bounds[Power](
                lower=bounds.lower, upper=bounds.upper
            ),
            _exclusion_bounds=system_bounds.exclusion_bounds,
        )

    @override
    def drop_old_proposals(self, loop_time: float) -> None:
        """Drop old proposals.

        This will remove all proposals that have not been updated for longer than
        `max_proposal_age`.

        Args:
            loop_time: The current loop time.
        """
        buckets_to_delete: list[frozenset[int]] = []
        for component_ids, proposals in self._component_buckets.items():
            to_delete: list[Proposal] = []
            for proposal in proposals:
                if (loop_time - proposal.creation_time) > self._max_proposal_age_sec:
                    to_delete.append(proposal)
            for proposal in to_delete:
                proposals.remove(proposal)
            if not proposals:
                buckets_to_delete.append(component_ids)

        for component_ids in buckets_to_delete:
            del self._component_buckets[component_ids]
            _ = self._target_power.pop(component_ids, None)



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Read and update logging severity from config."""

import logging
from dataclasses import dataclass, field
from typing import Annotated, Sequence, assert_never

import marshmallow
import marshmallow.validate

from ..actor import Actor
from ._manager import ConfigManager, wait_for_first

_logger = logging.getLogger(__name__)

LogLevel = Annotated[
    str,
    marshmallow.fields.String(
        validate=marshmallow.validate.OneOf(choices=logging.getLevelNamesMapping())
    ),
]
"""A marshmallow field for validating log levels."""


@dataclass(frozen=True, kw_only=True)
class LoggerConfig:
    """A configuration for a logger."""

    level: LogLevel = field(
        default="NOTSET",
        metadata={
            "metadata": {
                "description": "Log level for the logger. Uses standard logging levels."
            },
        },
    )
    """The log level for the logger."""


@dataclass(frozen=True, kw_only=True)
class LoggingConfig:
    """A configuration for the logging system."""

    root_logger: LoggerConfig = field(
        default_factory=lambda: LoggerConfig(level="INFO"),
        metadata={
            "metadata": {
                "description": "Default default configuration for all loggers.",
            },
        },
    )
    """The default log level."""

    loggers: dict[str, LoggerConfig] = field(
        default_factory=dict,
        metadata={
            "metadata": {
                "description": "Configuration for a logger (the key is the logger name)."
            },
        },
    )
    """The list of loggers configurations."""


class LoggingConfigUpdatingActor(Actor):
    """Actor that listens for logging configuration changes and sets them.

    Example:
        `config.toml` file:
        ```toml
        [logging.root_logger]
        level = "INFO"

        [logging.loggers."frequenz.sdk.actor.power_distributing"]
        level = "DEBUG"

        [logging.loggers."frequenz.channels"]
        level = "DEBUG"
        ```

        ```python
        import asyncio

        from frequenz.sdk.config import LoggingConfigUpdatingActor
        from frequenz.sdk.actor import run as run_actors

        async def run() -> None:
            config_manager: ConfigManager = ...
            await run_actors(LoggingConfigUpdatingActor(config_manager))

        asyncio.run(run())
        ```

        Now whenever the `config.toml` file is updated, the logging configuration
        will be updated as well.
    """

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        config_manager: ConfigManager,
        /,
        *,
        config_key: str | Sequence[str] = "logging",
        log_datefmt: str = "%Y-%m-%dT%H:%M:%S%z",
        log_format: str = "%(asctime)s %(levelname)-8s %(name)s:%(lineno)s: %(message)s",
        name: str | None = None,
    ):
        """Initialize this instance.

        Args:
            config_manager: The configuration manager to use.
            config_key: The key to use to retrieve the configuration from the
                configuration manager.  If `None`, the whole configuration will be used.
            log_datefmt: Use the specified date/time format in logs.
            log_format: Use the specified format string in logs.
            name: The name of this actor. If `None`, `str(id(self))` will be used. This
                is used mostly for debugging purposes.

        Note:
            The `log_format` and `log_datefmt` parameters are used in a call to
            `logging.basicConfig()`. If logging has already been configured elsewhere
            in the application (through a previous `basicConfig()` call), then the format
            settings specified here will be ignored.
        """
        self._config_receiver = config_manager.new_receiver(
            config_key, LoggingConfig, base_schema=None
        )

        # Setup default configuration.
        # This ensures logging is configured even if actor fails to start or
        # if the configuration cannot be loaded.
        self._current_config: LoggingConfig = LoggingConfig()

        super().__init__(name=name)

        logging.basicConfig(
            format=log_format,
            datefmt=log_datefmt,
            level=logging.INFO,
        )
        _logger.info("Applying initial default logging configuration...")
        self._reconfigure(self._current_config)

    async def _run(self) -> None:
        """Listen for configuration changes and update logging."""
        self._reconfigure(
            await wait_for_first(
                self._config_receiver, receiver_name=str(self), allow_none=True
            )
        )
        async for config_update in self._config_receiver:
            self._reconfigure(config_update)

    def _reconfigure(self, config_update: LoggingConfig | Exception | None) -> None:
        """Update the logging configuration.

        Args:
            config_update: The new configuration, or an exception if there was an error
                parsing the configuration, or `None` if the configuration was unset.
        """
        match config_update:
            case LoggingConfig():
                _logger.info(
                    "New configuration received, updating logging configuration."
                )
                self._update_logging(config_update)
            case None:
                _logger.info(
                    "Configuration was unset, resetting to the default "
                    "logging configuration."
                )
                self._update_logging(LoggingConfig())
            case Exception():
                _logger.info(
                    "New configuration has errors, keeping the old logging "
                    "configuration."
                )
            case unexpected:
                assert_never(unexpected)

    def _update_logging(self, config: LoggingConfig) -> None:
        """Configure the logging level."""
        # If the logger is not in the new config, set it to NOTSET
        loggers_to_unset = self._current_config.loggers.keys() - config.loggers.keys()
        for logger_id in loggers_to_unset:
            _logger.debug("Unsetting log level for logger '%s'", logger_id)
            logging.getLogger(logger_id).setLevel(logging.NOTSET)

        self._current_config = config
        _logger.info(
            "Setting root logger level to '%s'", self._current_config.root_logger.level
        )
        logging.getLogger().setLevel(self._current_config.root_logger.level)

        # For each logger in the new config, set the log level
        for logger_id, logger_config in self._current_config.loggers.items():
            _logger.info(
                "Setting log level for logger '%s' to '%s'",
                logger_id,
                logger_config.level,
            )
            logging.getLogger(logger_id).setLevel(logger_config.level)

        _logger.info("Logging config update completed.")



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Management of configuration."""

import asyncio
import logging
import pathlib
from collections.abc import Mapping, Sequence
from dataclasses import is_dataclass
from datetime import timedelta
from typing import Any, Final, Literal, TypeGuard, overload

import marshmallow
from frequenz.channels import Broadcast, Receiver, ReceiverStoppedError
from frequenz.channels.experimental import WithPrevious
from marshmallow import Schema, ValidationError
from typing_extensions import override

from ..actor._background_service import BackgroundService
from ._base_schema import BaseConfigSchema
from ._managing_actor import ConfigManagingActor
from ._util import DataclassT, _validate_load_kwargs, load_config

_logger = logging.getLogger(__name__)


class InvalidValueForKeyError(ValueError):
    """An error indicating that the value under the specified key is invalid."""

    def __init__(self, msg: str, *, key: Sequence[str], value: Any) -> None:
        """Initialize this error.

        Args:
            msg: The error message.
            key: The key that has an invalid value.
            value: The actual value that was found that is not a mapping.
        """
        super().__init__(msg)

        self.key: Final[Sequence[str]] = key
        """The key that has an invalid value."""

        self.value: Final[Any] = value
        """The actual value that was found that is not a mapping."""


class ConfigManager(BackgroundService):
    """A manager for configuration files.

    This class reads configuration files and sends the configuration to the receivers,
    providing configuration key filtering and value validation.

    For a more in-depth introduction and examples, please read the [module
    documentation][frequenz.sdk.config].
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        config_paths: str | pathlib.Path | Sequence[pathlib.Path | str],
        /,
        *,
        force_polling: bool = True,
        logging_config_key: str | Sequence[str] | None = "logging",
        name: str | None = None,
        polling_interval: timedelta = timedelta(seconds=1),
    ) -> None:
        """Initialize this config manager.

        Args:
            config_paths: The paths to the TOML files with the configuration. Order
                matters, as the configuration will be read and updated in the order
                of the paths, so the last path will override the configuration set by
                the previous paths. Dict keys will be merged recursively, but other
                objects (like lists) will be replaced by the value in the last path.
            force_polling: Whether to force file polling to check for changes.
            logging_config_key: The key to use for the logging configuration. If `None`,
                logging configuration will not be managed.  If a key is provided, the
                manager update the logging configuration whenever the configuration
                changes.
            name: A name to use when creating actors. If `None`, `str(id(self))` will
                be used. This is used mostly for debugging purposes.
            polling_interval: The interval to poll for changes. Only relevant if
                polling is enabled.
        """
        super().__init__(name=name)

        self.config_channel: Final[Broadcast[Mapping[str, Any]]] = Broadcast(
            name=f"{self}_config", resend_latest=True
        )
        """The channel used for sending configuration updates (resends the latest value).

        This is the channel used to communicate with the
        [`ConfigManagingActor`][frequenz.sdk.config.ConfigManager.config_actor] and will
        receive the complete raw configuration as a mapping.
        """

        self.config_actor: Final[ConfigManagingActor] = ConfigManagingActor(
            config_paths,
            self.config_channel.new_sender(),
            name=self.name,
            force_polling=force_polling,
            polling_interval=polling_interval,
        )
        """The actor that manages the configuration for this manager."""

        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from ._logging_actor import LoggingConfigUpdatingActor

        self.logging_actor: Final[LoggingConfigUpdatingActor | None] = (
            None
            if logging_config_key is None
            else LoggingConfigUpdatingActor(
                self, config_key=logging_config_key, name=self.name
            )
        )
        """The actor that manages the logging configuration for this manager."""

    @override
    def start(self) -> None:
        """Start this config manager."""
        self.config_actor.start()
        if self.logging_actor:
            self.logging_actor.start()

    @property
    @override
    def is_running(self) -> bool:
        """Whether this config manager is running."""
        return self.config_actor.is_running or (
            self.logging_actor is not None and self.logging_actor.is_running
        )

    @override
    def cancel(self, msg: str | None = None) -> None:
        """Cancel all running tasks and actors spawned by this config manager.

        Args:
            msg: The message to be passed to the tasks being cancelled.
        """
        if self.logging_actor:
            self.logging_actor.cancel(msg)
        self.config_actor.cancel(msg)

    @override
    async def wait(self) -> None:
        """Wait this config manager to finish.

        Wait until all tasks and actors are finished.

        Raises:
            BaseExceptionGroup: If any of the tasks spawned by this service raised an
                exception (`CancelError` is not considered an error and not returned in
                the exception group).
        """
        exceptions: list[BaseException] = []
        if self.logging_actor:
            try:
                await self.logging_actor
            except BaseExceptionGroup as err:  # pylint: disable=try-except-raise
                exceptions.append(err)

        try:
            await self.config_actor
        except BaseExceptionGroup as err:  # pylint: disable=try-except-raise
            exceptions.append(err)

        if exceptions:
            raise BaseExceptionGroup(f"Error while stopping {self!r}", exceptions)

    @override
    def __repr__(self) -> str:
        """Return a string representation of this config manager."""
        logging_actor = (
            f"logging_actor={self.logging_actor!r}, " if self.logging_actor else ""
        )
        return (
            f"<{self.__class__.__name__}: "
            f"name={self.name!r}, "
            f"config_channel={self.config_channel!r}, "
            + logging_actor
            + f"config_actor={self.config_actor!r}>"
        )

    def new_receiver(  # pylint: disable=too-many-arguments
        self,
        # This is tricky, because a str is also a Sequence[str], if we would use only
        # Sequence[str], then a regular string would also be accepted and taken as
        # a sequence, like "key" -> ["k", "e", "y"]. We should never remove the str from
        # the allowed types without changing Sequence[str] to something more specific,
        # like list[str] or tuple[str] (but both have their own problems).
        key: str | Sequence[str],
        config_class: type[DataclassT],
        /,
        *,
        skip_unchanged: bool = True,
        base_schema: type[Schema] | None = BaseConfigSchema,
        marshmallow_load_kwargs: dict[str, Any] | None = None,
    ) -> Receiver[DataclassT | Exception | None]:
        """Create a new receiver for receiving the configuration for a particular key.

        This method has a lot of features and functionalities to make it easier to
        receive configurations, but it also imposes some restrictions on how the
        configurations are received. If you need more control over the configuration
        receiver, you can create a receiver directly using
        [`config_channel.new_receiver()`][frequenz.sdk.config.ConfigManager.config_channel].

        For a more in-depth introduction and examples, please read the [module
        documentation][frequenz.sdk.config].

        Args:
            key: The configuration key to be read by the receiver. If a sequence of
                strings is used, it is used as a sub-key.
            config_class: The class object to use to instantiate a configuration. The
                configuration will be validated against this type too using
                [`marshmallow_dataclass`][].
            skip_unchanged: Whether to skip sending the configuration if it hasn't
                changed compared to the last one received.
            base_schema: An optional class to be used as a base schema for the
                configuration class. This allow using custom fields for example. Will be
                passed to [`marshmallow_dataclass.class_schema`][].
            marshmallow_load_kwargs: Additional arguments to be passed to
                [`marshmallow.Schema.load`][].

        Returns:
            The receiver for the configuration.
        """
        _validate_load_kwargs(marshmallow_load_kwargs)

        # We disable warning on overflow, because we are only interested in the latest
        # configuration, it is completely fine to drop old configuration updates.
        receiver = self.config_channel.new_receiver(
            name=f"{self}:{key}", limit=1, warn_on_overflow=False
        ).map(
            lambda config: _load_config_with_logging_and_errors(
                config,
                config_class,
                key=key,
                base_schema=base_schema,
                marshmallow_load_kwargs=marshmallow_load_kwargs,
            )
        )

        if skip_unchanged:
            # For some reason the type argument for WithPrevious is not inferred
            # correctly, so we need to specify it explicitly.
            return receiver.filter(
                WithPrevious[DataclassT | Exception | None](
                    lambda old, new: _not_equal_with_logging(
                        key=key, old_value=old, new_value=new
                    )
                )
            )

        return receiver


@overload
async def wait_for_first(
    receiver: Receiver[DataclassT | Exception | None],
    /,
    *,
    receiver_name: str | None = None,
    allow_none: Literal[False] = False,
    timeout: timedelta = timedelta(minutes=1),
) -> DataclassT: ...


@overload
async def wait_for_first(
    receiver: Receiver[DataclassT | Exception | None],
    /,
    *,
    receiver_name: str | None = None,
    allow_none: Literal[True] = True,
    timeout: timedelta = timedelta(minutes=1),
) -> DataclassT | None: ...


async def wait_for_first(
    receiver: Receiver[DataclassT | Exception | None],
    /,
    *,
    receiver_name: str | None = None,
    allow_none: bool = False,
    timeout: timedelta = timedelta(minutes=1),
) -> DataclassT | None:
    """Wait for and receive the the first configuration.

    For a more in-depth introduction and examples, please read the [module
    documentation][frequenz.sdk.config].

    Args:
        receiver: The receiver to receive the first configuration from.
        receiver_name: The name of the receiver, used for logging. If `None`, the
            string representation of the receiver will be used.
        allow_none: Whether consider a `None` value as a valid configuration.
        timeout: The timeout in seconds to wait for the first configuration.

    Returns:
        The first configuration received.

    Raises:
        asyncio.TimeoutError: If the first configuration is not received within the
            timeout.
        ReceiverStoppedError: If the receiver is stopped before the first configuration
            is received.
    """
    if receiver_name is None:
        receiver_name = str(receiver)

    # We need this type guard because we can't use a TypeVar for isinstance checks or
    # match cases.
    def is_config_class(value: DataclassT | Exception | None) -> TypeGuard[DataclassT]:
        return is_dataclass(value) if value is not None else False

    _logger.info(
        "%s: Waiting %s seconds for the first configuration to arrive...",
        receiver_name,
        timeout.total_seconds(),
    )
    try:
        async with asyncio.timeout(timeout.total_seconds()):
            async for config in receiver:
                match config:
                    case None:
                        if allow_none:
                            return None
                        _logger.error(
                            "%s: Received empty configuration, waiting again for "
                            "a first configuration to be set.",
                            receiver_name,
                        )
                    case Exception() as error:
                        _logger.error(
                            "%s: Error while receiving the first configuration, "
                            "will keep waiting for an update: %s.",
                            receiver_name,
                            error,
                        )
                    case config if is_config_class(config):
                        _logger.info("%s: Received first configuration.", receiver_name)
                        return config
                    case unexpected:
                        assert (
                            False
                        ), f"{receiver_name}: Unexpected value received: {unexpected!r}."
    except asyncio.TimeoutError:
        _logger.error("%s: No configuration received in time.", receiver_name)
        raise
    raise ReceiverStoppedError(receiver)


def _not_equal_with_logging(
    *,
    key: str | Sequence[str],
    old_value: DataclassT | Exception | None,
    new_value: DataclassT | Exception | None,
) -> bool:
    """Return whether the two mappings are not equal, logging if they are the same."""
    if old_value == new_value:
        _logger.info("Configuration has not changed for key %r, skipping update.", key)
        return False

    if isinstance(new_value, InvalidValueForKeyError) and not isinstance(
        old_value, InvalidValueForKeyError
    ):
        subkey_str = ""
        if key != new_value.key:
            subkey_str = f"When looking for sub-key {key!r}: "
        _logger.error(
            "%sConfiguration for key %r has an invalid value: %r",
            subkey_str,
            new_value.key,
            new_value.value,
        )
    return True


def _load_config_with_logging_and_errors(
    config: Mapping[str, Any],
    config_class: type[DataclassT],
    *,
    key: str | Sequence[str],
    base_schema: type[Schema] | None = None,
    marshmallow_load_kwargs: dict[str, Any] | None = None,
) -> DataclassT | Exception | None:
    """Load the configuration for the specified key, logging errors and returning them."""
    try:
        sub_config = _get_key(config, key)
        if sub_config is None:
            _logger.debug("Configuration key %r not found, sending None", key)
            return None

        loaded_config = _load_config(
            sub_config,
            config_class,
            key=key,
            base_schema=base_schema,
            marshmallow_load_kwargs=marshmallow_load_kwargs,
        )
        _logger.debug("Received new configuration: %s", loaded_config)
        return loaded_config
    except InvalidValueForKeyError as error:
        if len(key) > 1 and key != error.key:
            _logger.error("Error when looking for sub-key %r: %s", key, error)
        else:
            _logger.error(str(error))
        return error
    except ValidationError as error:
        _logger.error("The configuration for key %r is invalid: %s", key, error)
        return error
    except Exception as error:  # pylint: disable=broad-except
        _logger.exception(
            "An unexpected error occurred while loading the configuration for key %r: %s",
            key,
            error,
        )
        return error


def _get_key(
    config: Mapping[str, Any],
    # This is tricky, because a str is also a Sequence[str], if we would use only
    # Sequence[str], then a regular string would also be accepted and taken as
    # a sequence, like "key" -> ["k", "e", "y"]. We should never remove the str from
    # the allowed types without changing Sequence[str] to something more specific,
    # like list[str] or tuple[str].
    key: str | Sequence[str],
) -> Mapping[str, Any] | None:
    """Get the value from the configuration under the specified key.

    Args:
        config: The configuration to get the value from.
        key: The key to get the value for.

    Returns:
        The value under the key, or `None` if the key is not found.

    Raises:
        InvalidValueForKeyError: If the value under the key is not a mapping.
    """
    # We first normalize to a Sequence[str] to make it easier to work with.
    if isinstance(key, str):
        key = (key,)
    value = config
    current_path = []
    for subkey in key:
        current_path.append(subkey)
        if value is None:
            return None
        match value.get(subkey):
            case None:
                return None
            case Mapping() as new_value:
                value = new_value
            case invalid_value:
                raise InvalidValueForKeyError(
                    f"Value for key {current_path!r} is not a mapping: {invalid_value!r}",
                    key=current_path,
                    value=invalid_value,
                )
        value = new_value
    return value


def _load_config(
    config: Mapping[str, Any],
    config_class: type[DataclassT],
    *,
    key: str | Sequence[str],
    base_schema: type[Schema] | None = BaseConfigSchema,
    marshmallow_load_kwargs: dict[str, Any] | None = None,
) -> DataclassT | InvalidValueForKeyError | ValidationError | None:
    """Try to load a configuration and log any validation errors."""
    loaded_config = load_config(
        config_class,
        config,
        base_schema=base_schema,
        marshmallow_load_kwargs=marshmallow_load_kwargs,
    )

    marshmallow_load_kwargs = (
        {} if marshmallow_load_kwargs is None else marshmallow_load_kwargs.copy()
    )

    # When excluding unknown fields we still want to notify the user, as
    # this could mean there is a typo in the configuration and some value is
    # not being loaded as desired.
    marshmallow_load_kwargs["unknown"] = marshmallow.RAISE
    try:
        load_config(
            config_class,
            config,
            base_schema=base_schema,
            marshmallow_load_kwargs=marshmallow_load_kwargs,
        )
    except ValidationError as err:
        _logger.warning(
            "The configuration for key %r has extra fields that will be ignored: %s",
            key,
            err,
        )

    return loaded_config



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Configuration management.

# Overview

To provide dynamic configurations to an application, you can use the
[`ConfigManager`][frequenz.sdk.config.ConfigManager] class. This class provides
a convenient interface to manage configurations from multiple config files and receive
updates when the configurations change.  Users can create a receiver to receive
configurations from the manager.

# Setup

To use the `ConfigManager`, you need to create an instance of it and pass the
paths to the configuration files. The configuration files must be in the TOML
format.

When specifying multiple files order matters, as the configuration will be read and
updated in the order of the paths, so the last path will override the configuration set
by the previous paths. Dict keys will be merged recursively, but other objects (like
lists) will be replaced by the value in the last path.

```python
from frequenz.sdk.config import ConfigManager

async with ConfigManager(["base-config.toml", "overrides.toml"]) as config_manager:
    ...
```

# Logging

The `ConfigManager` can also instantiate
a [`LoggingConfigUpdatingActor`][frequenz.sdk.config.LoggingConfigUpdatingActor] to
monitor logging configurations. This actor will listen for logging configuration changes
and update the logging configuration accordingly.

This feature is enabled by default using the key `logging` in the configuration file. To
disable it you can pass `logging_config_key=None` to the `ConfigManager`.

# Receiving configurations

To receive configurations, you can create a receiver using the [`new_receiver()`][
frequenz.sdk.config.ConfigManager.new_receiver] method. The receiver will receive
configurations from the manager for a particular key, and validate and load the
configurations to a dataclass using [`marshmallow_dataclass`][].

If the key is a sequence of strings, it will be treated as a nested key and the
receiver will receive the configuration under the nested key. For example
`["key", "subkey"]` will get only `config["key"]["subkey"]`.

Besides a configuration instance, the receiver can also receive exceptions if there are
errors loading the configuration (typically
a [`ValidationError`][marshmallow.ValidationError]), or `None` if there is no
configuration for the key.

The value under `key` must be another mapping, otherwise
a [`InvalidValueForKeyError`][frequenz.sdk.config.InvalidValueForKeyError] instance will
be sent to the receiver.

If there were any errors loading the configuration, the error will be logged too.

```python
from dataclasses import dataclass
from frequenz.sdk.config import ConfigManager

@dataclass(frozen=True, kw_only=True)
class AppConfig:
    test: int

async with ConfigManager("config.toml") as config_manager:
    receiver = config_manager.new_receiver("app", AppConfig)
    app_config = await receiver.receive()
    match app_config:
        case AppConfig(test=42):
            print("App configured with 42")
        case Exception() as error:
            print(f"Error loading configuration: {error}")
        case None:
            print("There is no configuration for the app key")
```

## Validation and loading

The configuration class used to create the configuration instance is expected to be
a [`dataclasses.dataclass`][], which is used to create a [`marshmallow.Schema`][] via
the [`marshmallow_dataclass.class_schema`][] function.

This means you can customize the schema derived from the configuration
dataclass using [`marshmallow_dataclass`][] to specify extra validation and
options via field metadata.

Customization can also be done via a `base_schema`. By default
[`BaseConfigSchema`][frequenz.sdk.config.BaseConfigSchema] is used to provide support
for some extra commonly used fields (like [quantities][frequenz.quantities]) and to
exclude unknown fields by default.

```python
import marshmallow.validate
from dataclasses import dataclass, field

@dataclass(frozen=True, kw_only=True)
class Config:
    test: int = field(
        metadata={"validate": marshmallow.validate.Range(min=0)},
    )
```

Additional arguments can be passed to [`marshmallow.Schema.load`][] using
the `marshmallow_load_kwargs` keyword arguments.

When [`marshmallow.EXCLUDE`][] is used, a warning will be logged if there are extra
fields in the configuration that are excluded. This is useful, for example, to catch
typos in the configuration file.

## Skipping superfluous updates

If there is a burst of configuration updates, the receiver will only receive the
last configuration, older configurations will be ignored.

If `skip_unchanged` is set to `True`, then a configuration that didn't change
compared to the last one received will be ignored and not sent to the receiver.
The comparison is done using the *raw* `dict` to determine if the configuration
has changed.

## Error handling

The value under `key` must be another mapping, otherwise an error
will be logged and a [`frequenz.sdk.config.InvalidValueForKeyError`][] instance
will be sent to the receiver.

Configurations that don't pass the validation will be logged as an error and
the [`ValidationError`][marshmallow.ValidationError] sent to the receiver.

Any other unexpected error raised during the configuration loading will be
logged as an error and the error instance sent to the receiver.

## Further customization

If you have special needs for receiving the configurations (for example validating using
`marshmallow` doesn't fit your needs), you can create a custom receiver using
[`config_channel.new_receiver()`][frequenz.sdk.config.ConfigManager.config_channel]
directly. Please bear in mind that this provides a low-level access to the whole config
in the file as a raw Python mapping.

# Recommended usage

Actors that need to be reconfigured should take a configuration manager and a key to
receive configurations updates, and instantiate the new receiver themselves. This allows
actors to have full control over how the configuration is loaded (for example providing
a custom base schema or marshmallow options).

Passing the key explicitly too allows application to structure the configuration in
whatever way is most convenient for the application.

Actors can use the [`wait_for_first()`][frequenz.sdk.config.wait_for_first] function to
wait for the first configuration to be received, and cache the configuration for later
use and in case the actor is restarted. If the configuration is not received after some
timeout, a [`asyncio.TimeoutError`][] will be raised (and if uncaught, the actor will
be automatically restarted after some delay).

Example: Actor that can run without a configuration (using a default configuration)
    ```python title="actor.py" hl_lines="18 34 42 62 64"
    import dataclasses
    import logging
    from collections.abc import Sequence
    from datetime import timedelta
    from typing import assert_never

    from frequenz.channels import select, selected_from
    from frequenz.channels.event import Event

    from frequenz.sdk.actor import Actor
    from frequenz.sdk.config import ConfigManager, wait_for_first

    _logger = logging.getLogger(__name__)

    @dataclasses.dataclass(frozen=True, kw_only=True)
    class MyActorConfig:
        some_config: timedelta = dataclasses.field(
            default=timedelta(seconds=42), # (1)!
            metadata={"metadata": {"description": "Some optional configuration"}},
        )

    class MyActor(Actor):
        def __init__(
            self,
            config_manager: ConfigManager,
            /,
            *,
            config_key: str | Sequence[str],
            name: str | None = None,
        ) -> None:
            super().__init__(name=name)
            self._config_manager = config_manager
            self._config_key = config_key
            self._config: MyActorConfig = MyActorConfig() # (2)!

        async def _run(self) -> None:
            config_receiver = self._config_manager.new_receiver(
                self._config_key, MyActorConfig
            )
            self._update_config(
                await wait_for_first(
                    config_receiver, receiver_name=str(self), allow_none=True # (3)!
                )
            )

            other_receiver = Event()

            async for selected in select(config_receiver, other_receiver):
                if selected_from(selected, config_receiver):
                    self._update_config(selected.message)
                elif selected_from(selected, other_receiver):
                    # Do something else
                    ...

        def _update_config(self, config_update: MyActorConfig | Exception | None) -> None:
            match config_update:
                case MyActorConfig() as config:
                    _logger.info("New configuration received, updating.")
                    self._reconfigure(config)
                case None:
                    _logger.info("Configuration was unset, resetting to the default")
                    self._reconfigure(MyActorConfig()) # (4)!
                case Exception():
                    _logger.info( # (5)!
                        "New configuration has errors, keeping the old configuration."
                    )
                case unexpected:
                    assert_never(unexpected)

        def _reconfigure(self, config: MyActorConfig) -> None:
            self._config = config
            # Do something with the new configuration
    ```

    1. This is different when the actor requires a configuration to run. Here, the
        config has a default value.
    2. This is different when the actor requires a configuration to run. Here, the actor
        can just instantiate a default configuration.
    3. This is different when the actor requires a configuration to run. Here, the actor
        can accept a `None` configuration.
    4. This is different when the actor requires a configuration to run. Here, the actor
        can reset to a default configuration.
    5. There is no need to log the error itself, the configuration manager will log it
        automatically.

Example: Actor that requires a configuration to run
    ```python title="actor.py" hl_lines="17 33 40 58 60"
    import dataclasses
    import logging
    from collections.abc import Sequence
    from datetime import timedelta
    from typing import assert_never

    from frequenz.channels import select, selected_from
    from frequenz.channels.event import Event

    from frequenz.sdk.actor import Actor
    from frequenz.sdk.config import ConfigManager, wait_for_first

    _logger = logging.getLogger(__name__)

    @dataclasses.dataclass(frozen=True, kw_only=True)
    class MyActorConfig:
        some_config: timedelta = dataclasses.field( # (1)!
            metadata={"metadata": {"description": "Some required configuration"}},
        )

    class MyActor(Actor):
        def __init__(
            self,
            config_manager: ConfigManager,
            /,
            *,
            config_key: str | Sequence[str],
            name: str | None = None,
        ) -> None:
            super().__init__(name=name)
            self._config_manager = config_manager
            self._config_key = config_key
            self._config: MyActorConfig # (2)!

        async def _run(self) -> None:
            config_receiver = self._config_manager.new_receiver(
                self._config_key, MyActorConfig
            )
            self._update_config(
                await wait_for_first(config_receiver, receiver_name=str(self)) # (3)!
            )

            other_receiver = Event()

            async for selected in select(config_receiver, other_receiver):
                if selected_from(selected, config_receiver):
                    self._update_config(selected.message)
                elif selected_from(selected, other_receiver):
                    # Do something else
                    ...

        def _update_config(self, config_update: MyActorConfig | Exception | None) -> None:
            match config_update:
                case MyActorConfig() as config:
                    _logger.info("New configuration received, updating.")
                    self._reconfigure(config)
                case None:
                    _logger.info("Configuration was unset, keeping the old configuration.") # (4)!
                case Exception():
                    _logger.info( # (5)!
                        "New configuration has errors, keeping the old configuration."
                    )
                case unexpected:
                    assert_never(unexpected)

        def _reconfigure(self, config: MyActorConfig) -> None:
            self._config = config
            # Do something with the new configuration
    ```

    1. This is different when the actor can use a default configuration. Here, the
        field is required, so there is no default configuration possible.
    2. This is different when the actor can use a default configuration. Here, the
       assignment of the configuration is delayed to the `_run()` method.
    3. This is different when the actor can use a default configuration. Here, the actor
        doesn't accept `None` as a valid configuration as it can't create a default
        configuration.
    4. This is different when the actor can use a default configuration. Here, the actor
        doesn't accept `None` as a valid configuration as it can't create a default
        configuration, so it needs to keep the old configuration.
    5. There is no need to log the error itself, the configuration manager will log it
        automatically.


Example: Application
    The pattern used by the application is very similar to the one used by actors. In
    this case the application requires a configuration to run, but if it could also use
    a default configuration, the changes would be the same as in the actor examples.

    ```python title="app.py" hl_lines="14"
    import asyncio
    import dataclasses
    import logging
    import pathlib
    from collections.abc import Sequence
    from datetime import timedelta
    from typing import Sequence, assert_never

    from frequenz.sdk.actor import Actor
    from frequenz.sdk.config import ConfigManager, wait_for_first

    _logger = logging.getLogger(__name__)

    class MyActor(Actor): # (1)!
        def __init__(
            self, config_manager: ConfigManager, /, *, config_key: str | Sequence[str]
        ) -> None:
            super().__init__()
            self._config_manager = config_manager
            self._config_key = config_key
        async def _run(self) -> None: ...

    @dataclasses.dataclass(frozen=True, kw_only=True)
    class AppConfig:
        enable_actor: bool = dataclasses.field(
            metadata={"metadata": {"description": "Whether to enable the actor"}},
        )

    class App:
        def __init__(self, *, config_paths: Sequence[pathlib.Path]):
            self._config_manager = ConfigManager(config_paths)
            self._config_receiver = self._config_manager.new_receiver("app", AppConfig)
            self._actor = MyActor(self._config_manager, config_key="actor")

        async def _update_config(self, config_update: AppConfig | Exception | None) -> None:
            match config_update:
                case AppConfig() as config:
                    _logger.info("New configuration received, updating.")
                    await self._reconfigure(config)
                case None:
                    _logger.info("Configuration was unset, keeping the old configuration.")
                case Exception():
                    _logger.info("New configuration has errors, keeping the old configuration.")
                case unexpected:
                    assert_never(unexpected)

        async def _reconfigure(self, config: AppConfig) -> None:
            if config.enable_actor:
                self._actor.start()
            else:
                await self._actor.stop()

        async def run(self) -> None:
            _logger.info("Starting App...")

            async with self._config_manager:
                await self._update_config(
                    await wait_for_first(self._config_receiver, receiver_name="app")
                )

                _logger.info("Waiting for configuration updates...")
                async for config_update in self._config_receiver:
                    await self._reconfigure(config_update)

    if __name__ == "__main__":
        asyncio.run(App(config_paths="config.toml").run())
    ```

    1. Look for the actor examples for a proper implementation of the actor.

    Example configuration file:

    ```toml title="config.toml"
    [app]
    enable_actor = true

    [actor]
    some_config = 10

    [logging.root_logger]
    level = "DEBUG"
    ```
"""

from ._base_schema import BaseConfigSchema
from ._logging_actor import LoggerConfig, LoggingConfig, LoggingConfigUpdatingActor
from ._manager import ConfigManager, InvalidValueForKeyError, wait_for_first
from ._managing_actor import ConfigManagingActor
from ._util import load_config

__all__ = [
    "BaseConfigSchema",
    "ConfigManager",
    "ConfigManagingActor",
    "InvalidValueForKeyError",
    "LoggerConfig",
    "LoggingConfig",
    "LoggingConfigUpdatingActor",
    "load_config",
    "wait_for_first",
]



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Base schema for configuration classes."""

import marshmallow
from frequenz.quantities.experimental.marshmallow import QuantitySchema


class BaseConfigSchema(QuantitySchema):
    """A base schema for configuration classes.

    This schema provides validation for quantities and ignores unknown fields by
    default.
    """

    class Meta:
        """Meta options for the schema."""

        unknown = marshmallow.EXCLUDE



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Load shedding actor."""

import asyncio
import random
import sys
import termios
import tty
from dataclasses import dataclass
from datetime import datetime, timezone
from heapq import heappop, heappush
from typing import AsyncGenerator
from unittest.mock import patch

from frequenz.channels import Broadcast, Receiver, Sender
from frequenz.quantities import Percentage, Power

from frequenz.sdk import microgrid
from frequenz.sdk.actor import Actor, run
from frequenz.sdk.timeseries import Sample

# Mock configuration
CONF_STATE = {}


def mock_set_consumer(name: str, power: float) -> None:
    """Mock setting consumer power by storing the state in a dictionary.

    Args:
        name: Consumer name.
        power: Power value to set.
    """
    CONF_STATE[name] = power


def _log(msg: str) -> None:
    print(msg, end="\n\r")


class PowerMockActor(Actor):
    """Power Mock Actor.

    Asynchronously listens to user key presses 'm' and 'n' to increase and decrease power of a
    static consumer.
    """

    def __init__(self, consumer_name: str) -> None:
        """Initialize the actor."""
        super().__init__()
        self.consumer_name = consumer_name
        self.power_step = Power.from_kilowatts(1.0)

    async def _run(self) -> None:
        _log("Press 'm' to increase power or 'n' to decrease power for the consumer.")

        while True:
            # Call _read_key in a thread to avoid blocking the event loop
            key = await asyncio.to_thread(self._read_key)
            if key == "m":
                CONF_STATE[self.consumer_name] = (
                    CONF_STATE.get(self.consumer_name, 0) + self.power_step.as_watts()
                )
                _log(
                    f"Increased {self.consumer_name} power to "
                    f"{CONF_STATE[self.consumer_name]/1000.0} kW"
                )
            elif key == "n":
                CONF_STATE[self.consumer_name] = max(
                    0,
                    CONF_STATE.get(self.consumer_name, 0) - self.power_step.as_watts(),
                )
                _log(
                    f"Decreased {self.consumer_name} power to "
                    f"{CONF_STATE[self.consumer_name]/1000.0} kW"
                )
            elif key == "q":
                sys.exit()
            else:
                _log("Invalid key. Use 'm' or 'n'.")

    def _read_key(self) -> str:
        """Read a single key press without waiting for Enter."""
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            key = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return key


@dataclass(order=True)
class Consumer:
    """Consumer dataclass."""

    priority: int
    name: str
    power: Power
    enabled: bool = False


class LoadSheddingActor(Actor):
    """Simple load shedding actor."""

    def __init__(
        self,
        max_peak: Power,
        consumers: list[Consumer],
        grid_meter_receiver: Receiver[Sample[Power]],
    ):
        """Initialize the actor."""
        super().__init__()
        self.max_peak = max_peak
        self.disable_tolerance = self.max_peak * 0.9
        self.enable_tolerance = self.max_peak * 0.8
        self.grid_meter_receiver = grid_meter_receiver

        self.enabled_consumers: list[Consumer] = []
        self.disabled_consumers: list[Consumer] = []

        for c in consumers:
            heappush(self.disabled_consumers, c)

    async def _enable_consumer(self, consumer: Consumer) -> None:
        if not consumer.enabled:
            consumer.enabled = True
            heappush(self.enabled_consumers, consumer)
            _log(f"+++{consumer.name}, +{consumer.power}")
            # This is a mock function to set the consumer power,
            # in a real system this would be replaced with the actual implementation
            mock_set_consumer(consumer.name, consumer.power.as_watts())

    async def _disable_consumer(self, consumer: Consumer) -> None:
        if consumer.enabled:
            consumer.enabled = False
            heappush(self.disabled_consumers, consumer)
            _log(f"---{consumer.name}, -{consumer.power}")
            # This is a mock function to set the consumer power,
            # in a real system this would be replaced with the actual implementation
            mock_set_consumer(consumer.name, 0)

    async def _adjust_loads(self, current_load: Power) -> None:
        while current_load > self.disable_tolerance and self.enabled_consumers:
            enabled_consumer: Consumer = heappop(self.enabled_consumers)
            await self._disable_consumer(enabled_consumer)
            current_load -= enabled_consumer.power

        temp_disabled: list[Consumer] = []
        while self.disabled_consumers:
            disabled_consumer: Consumer = heappop(self.disabled_consumers)
            if current_load + disabled_consumer.power <= self.enable_tolerance:
                await self._enable_consumer(disabled_consumer)
                current_load += disabled_consumer.power
            else:
                heappush(temp_disabled, disabled_consumer)
                break

        while temp_disabled:
            heappush(self.disabled_consumers, heappop(temp_disabled))

    async def _run(self) -> None:
        async for power_sample in self.grid_meter_receiver:
            if power_sample.value:
                _log(
                    f"Power: {power_sample.value}, "
                    f"Peak: {self.max_peak} ({self.disable_tolerance} / {self.enable_tolerance})"
                    f", Enabled: {', '.join(c.name for c in self.enabled_consumers)}\r"
                )
                await self._adjust_loads(power_sample.value)


async def mock_sender(
    sender: Sender[Sample[Power]],
) -> AsyncGenerator[Sample[Power], None]:
    """Mock implementation of a grid meter receiver.

    It sends power values every second.
    """
    current_load = Power.from_kilowatts(0.0)

    def compute_power() -> Power:
        """Compute current grid power based on mock state."""
        return Power.from_watts(sum(CONF_STATE.values()))

    while True:
        current_load = compute_power()
        # Add +- 8% noise to the current load
        current_load += current_load * Percentage.from_fraction(
            random.uniform(-0.08, 0.08)
        )
        await sender.send(
            Sample(timestamp=datetime.now(tz=timezone.utc), value=current_load)
        )
        await asyncio.sleep(1)


async def main() -> None:
    """Program entry point."""
    consumers = [
        Consumer(priority=1, name="Fan2", power=Power.from_kilowatts(2.5)),
        Consumer(priority=2, name="Drier1", power=Power.from_kilowatts(3.0)),
        Consumer(priority=2, name="Drier2", power=Power.from_kilowatts(2.0)),
        Consumer(priority=3, name="Conveyor1", power=Power.from_kilowatts(1.5)),
        Consumer(priority=3, name="Conveyor2", power=Power.from_kilowatts(1.0)),
        Consumer(priority=4, name="Auger", power=Power.from_kilowatts(2.0)),
        Consumer(priority=4, name="HopperMixer", power=Power.from_kilowatts(2.5)),
        Consumer(priority=5, name="SiloVentilation", power=Power.from_kilowatts(1.0)),
        Consumer(priority=5, name="LoaderArm", power=Power.from_kilowatts(3.0)),
        Consumer(priority=6, name="SeedCleaner", power=Power.from_kilowatts(2.5)),
        Consumer(priority=6, name="Sprayer", power=Power.from_kilowatts(2.0)),
        Consumer(priority=7, name="Grinder", power=Power.from_kilowatts(3.0)),
        Consumer(priority=7, name="Shaker", power=Power.from_kilowatts(1.5)),
        Consumer(priority=8, name="Sorter", power=Power.from_kilowatts(2.0)),
    ]

    for consumer in consumers:
        mock_set_consumer(consumer.name, 0)

    grid_meter_receiver = microgrid.grid().power.new_receiver()

    actor_instance = LoadSheddingActor(
        max_peak=Power.from_kilowatts(30),
        consumers=consumers,
        grid_meter_receiver=grid_meter_receiver,
    )

    user_input_actor = PowerMockActor(consumer_name="static_consumer")

    await run(actor_instance, user_input_actor)


if __name__ == "__main__":
    with patch("frequenz.sdk.microgrid.grid") as mock_grid:
        chan = Broadcast[Sample[Power]](name="grid_power")
        mock_grid.return_value.power.new_receiver = chan.new_receiver

        async def begin() -> None:
            """Start main & mock sender."""
            await asyncio.gather(
                main(),
                mock_sender(chan.new_sender()),
            )

        asyncio.run(begin())



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""The power manager."""

from __future__ import annotations

import asyncio
import logging
import sys
from datetime import datetime, timedelta, timezone
from typing import assert_never

from frequenz.channels import Receiver, Sender, select, selected_from
from frequenz.channels.timer import SkipMissedAndDrift, Timer
from frequenz.client.microgrid import ComponentCategory, ComponentType, InverterType
from typing_extensions import override

from ..._internal._asyncio import run_forever
from ..._internal._channels import ChannelRegistry
from ...actor import Actor
from ...timeseries._base_types import SystemBounds
from .. import _data_pipeline, _power_distributing
from ._base_classes import Algorithm, BaseAlgorithm, Proposal, ReportRequest, _Report
from ._matryoshka import Matryoshka
from ._shifting_matryoshka import ShiftingMatryoshka

_logger = logging.getLogger(__name__)


class PowerManagingActor(Actor):
    """The power manager."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        proposals_receiver: Receiver[Proposal],
        bounds_subscription_receiver: Receiver[ReportRequest],
        power_distributing_requests_sender: Sender[_power_distributing.Request],
        power_distributing_results_receiver: Receiver[_power_distributing.Result],
        channel_registry: ChannelRegistry,
        algorithm: Algorithm,
        component_category: ComponentCategory,
        component_type: ComponentType | None = None,
    ):
        """Create a new instance of the power manager.

        Args:
            proposals_receiver: The receiver for proposals.
            bounds_subscription_receiver: The receiver for bounds subscriptions.
            power_distributing_requests_sender: The sender for power distribution
                requests.
            power_distributing_results_receiver: The receiver for power distribution
                results.
            channel_registry: The channel registry.
            algorithm: The power management algorithm to use.
            component_category: The category of the component this power manager
                instance is going to support.
            component_type: The type of the component of the given category that this
                actor is responsible for.  This is used only when the component category
                is not enough to uniquely identify the component.  For example, when the
                category is `ComponentCategory.INVERTER`, the type is needed to identify
                the inverter as a solar inverter or a battery inverter.  This can be
                `None` when the component category is enough to uniquely identify the
                component.
        """
        self._component_category = component_category
        self._component_type = component_type
        self._bounds_subscription_receiver = bounds_subscription_receiver
        self._power_distributing_requests_sender = power_distributing_requests_sender
        self._power_distributing_results_receiver = power_distributing_results_receiver
        self._channel_registry = channel_registry
        self._proposals_receiver = proposals_receiver

        self._system_bounds: dict[frozenset[int], SystemBounds] = {}
        self._bound_tracker_tasks: dict[frozenset[int], asyncio.Task[None]] = {}
        self._subscriptions: dict[frozenset[int], dict[int, Sender[_Report]]] = {}

        match algorithm:
            case Algorithm.MATRYOSHKA:
                self._algorithm: BaseAlgorithm = Matryoshka(
                    max_proposal_age=timedelta(seconds=60.0)
                )
            case Algorithm.SHIFTING_MATRYOSHKA:
                self._algorithm = ShiftingMatryoshka(
                    max_proposal_age=timedelta(seconds=60.0)
                )
            case _:
                assert_never(algorithm)

        super().__init__()

    async def _send_reports(self, component_ids: frozenset[int]) -> None:
        """Send reports for a set of components, to all subscribers.

        Args:
            component_ids: The component IDs for which a collective report should be
                sent.
        """
        bounds = self._system_bounds.get(component_ids)
        if bounds is None:
            _logger.warning("PowerManagingActor: No bounds for %s", component_ids)
            return
        for priority, sender in self._subscriptions.get(component_ids, {}).items():
            status = self._algorithm.get_status(
                component_ids,
                priority,
                bounds,
            )
            await sender.send(status)

    async def _bounds_tracker(
        self,
        component_ids: frozenset[int],
        bounds_receiver: Receiver[SystemBounds],
    ) -> None:
        """Track the power bounds of a set of components and update the cache.

        Args:
            component_ids: The component IDs for which this task should track the
                collective bounds of.
            bounds_receiver: The receiver for power bounds.
        """
        async for bounds in bounds_receiver:
            self._system_bounds[component_ids] = bounds
            await self._send_updated_target_power(component_ids, None)
            await self._send_reports(component_ids)

    def _add_system_bounds_tracker(self, component_ids: frozenset[int]) -> None:
        """Add a system bounds tracker for the given components.

        Args:
            component_ids: The component IDs for which to add a bounds tracker.

        Raises:
            NotImplementedError: When the pool type is not supported.
        """
        bounds_receiver: Receiver[SystemBounds]
        # pylint: disable=protected-access
        if self._component_category is ComponentCategory.BATTERY:
            battery_pool = _data_pipeline.new_battery_pool(
                priority=-sys.maxsize - 1, component_ids=component_ids
            )
            bounds_receiver = battery_pool._system_power_bounds.new_receiver()
        elif self._component_category is ComponentCategory.EV_CHARGER:
            ev_charger_pool = _data_pipeline.new_ev_charger_pool(
                priority=-sys.maxsize - 1, component_ids=component_ids
            )
            bounds_receiver = ev_charger_pool._system_power_bounds.new_receiver()
        elif (
            self._component_category is ComponentCategory.INVERTER
            and self._component_type is InverterType.SOLAR
        ):
            pv_pool = _data_pipeline.new_pv_pool(
                priority=-sys.maxsize - 1, component_ids=component_ids
            )
            bounds_receiver = pv_pool._system_power_bounds.new_receiver()
        # pylint: enable=protected-access
        else:
            err = (
                "PowerManagingActor: Unsupported component category: "
                f"{self._component_category}"
            )
            _logger.error(err)
            raise NotImplementedError(err)

        self._system_bounds[component_ids] = SystemBounds(
            timestamp=datetime.now(tz=timezone.utc),
            inclusion_bounds=None,
            exclusion_bounds=None,
        )

        # Start the bounds tracker, for ongoing updates.
        self._bound_tracker_tasks[component_ids] = asyncio.create_task(
            run_forever(lambda: self._bounds_tracker(component_ids, bounds_receiver))
        )

    async def _send_updated_target_power(
        self,
        component_ids: frozenset[int],
        proposal: Proposal | None,
        must_send: bool = False,
    ) -> None:
        target_power = self._algorithm.calculate_target_power(
            component_ids,
            proposal,
            self._system_bounds[component_ids],
            must_send,
        )
        if target_power is not None:
            await self._power_distributing_requests_sender.send(
                _power_distributing.Request(
                    power=target_power,
                    component_ids=component_ids,
                    adjust_power=True,
                )
            )

    @override
    async def _run(self) -> None:
        """Run the power managing actor."""
        last_result_partial_failure = False
        drop_old_proposals_timer = Timer(timedelta(seconds=1.0), SkipMissedAndDrift())
        async for selected in select(
            self._proposals_receiver,
            self._bounds_subscription_receiver,
            self._power_distributing_results_receiver,
            drop_old_proposals_timer,
        ):
            if selected_from(selected, self._proposals_receiver):
                proposal = selected.message
                if proposal.component_ids not in self._bound_tracker_tasks:
                    self._add_system_bounds_tracker(proposal.component_ids)

                # TODO: must_send=True forces a new request to # pylint: disable=fixme
                # be sent to the PowerDistributor, even if there's no change in power.
                #
                # This is needed because requests would expire in the microgrid service
                # otherwise.
                #
                # This can be removed as soon as
                # https://github.com/frequenz-floss/frequenz-sdk-python/issues/293 is
                # implemented.
                await self._send_updated_target_power(
                    proposal.component_ids, proposal, must_send=True
                )
                await self._send_reports(proposal.component_ids)

            elif selected_from(selected, self._bounds_subscription_receiver):
                sub = selected.message
                component_ids = sub.component_ids
                priority = sub.priority

                if component_ids not in self._subscriptions:
                    self._subscriptions[component_ids] = {
                        priority: self._channel_registry.get_or_create(
                            _Report, sub.get_channel_name()
                        ).new_sender()
                    }
                elif priority not in self._subscriptions[component_ids]:
                    self._subscriptions[component_ids][priority] = (
                        self._channel_registry.get_or_create(
                            _Report, sub.get_channel_name()
                        ).new_sender()
                    )

                if sub.component_ids not in self._bound_tracker_tasks:
                    self._add_system_bounds_tracker(sub.component_ids)

            elif selected_from(selected, self._power_distributing_results_receiver):
                result = selected.message
                if not isinstance(result, _power_distributing.Success):
                    _logger.warning(
                        "PowerManagingActor: PowerDistributing failed: %s", result
                    )
                match result:
                    case _power_distributing.PartialFailure(request):
                        if not last_result_partial_failure:
                            last_result_partial_failure = True
                            await self._send_updated_target_power(
                                frozenset(request.component_ids), None, must_send=True
                            )
                    case _power_distributing.Success():
                        last_result_partial_failure = False
                await self._send_reports(frozenset(result.request.component_ids))

            elif selected_from(selected, drop_old_proposals_timer):
                self._algorithm.drop_old_proposals(asyncio.get_event_loop().time())



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Class that stores values of the component metrics."""


from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime

from frequenz.client.microgrid import ComponentMetricId


@dataclass(frozen=True, eq=False)
class ComponentMetricsData:
    """Store values of the component metrics."""

    component_id: int
    """The component ID the data is for."""

    timestamp: datetime
    """The timestamp for all the metrics."""

    metrics: Mapping[ComponentMetricId, float]
    """The values for each metric."""

    def get(self, metric: ComponentMetricId) -> float | None:
        """Get metric value.

        Args:
            metric: metric id

        Returns:
            Value of the metric.
        """
        return self.metrics.get(metric, None)

    def __eq__(self, other: object) -> bool:
        """Compare two objects of this class.

        Object are considered as equal if all stored values except for timestamp
        are equal.

        Args:
            other: object to compare.

        Returns:
            True if two objects are equal, false otherwise.
        """
        if not isinstance(other, ComponentMetricsData):
            return False

        return self.component_id == other.component_id and self.metrics == other.metrics
